@@ -1,5 +1,5 @@
 # This file is part of bugzillatools
-# Copyright (C) 2011 Benon Technologies Pty Ltd
+# Copyright (C) 2011 Benon Technologies Pty Ltd, Fraser Tweedale
 #
 # bugzillatools is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -14,9 +14,11 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+import urlparse
 import xmlrpclib
 
 from . import bug
+from . import config
 
 
 # field type constants
@@ -34,13 +36,19 @@ class UserError(Exception):
     pass
 
 
+class URLError(Exception):
+    pass
+
+
 class Bugzilla(object):
 
-    _products = None
-    _fields = None
-    _user_cache = {}
+    __slots__ = [
+        '_products', '_fields', '_user_cache',
+        'url', 'user', 'password', 'config',
+        'server',
+    ]
 
-    def __init__(self, url, user, password):
+    def __init__(self, url=None, user=None, password=None, **config):
         """Create a Bugzilla XML-RPC client.
 
         url      : points to a bugzilla instance (base URL; must end in '/')
@@ -48,13 +56,29 @@ class Bugzilla(object):
         password : bugzilla password
         """
 
+        self._products = None
+        self._fields = None
+        self._user_cache = {}
+
+        self.url = url
         self.user = user
         self.password = password
+        self.config = config
 
-        # TODO URL sanity checks
-        url = url[:-1] if url[-1] == '/' else url  # strip trailing slash
+        parsed_url = urlparse.urlparse(url)
+        if not parsed_url.netloc:
+            raise URLError('URL {!r} is not valid.'.format(url))
+        if parsed_url.scheme not in ('http', 'https'):
+            raise URLError(
+                'URL scheme {!r} not supported.'.format(parsed_url.scheme)
+            )
+        if parsed_url.params or parsed_url.query or parsed_url.fragment:
+            raise URLError(
+                'URL params, queries and fragments not supported.'
+            )
+        url = url + 'xmlrpc.cgi' if url[-1] == '/' else url + '/xmlrpc.cgi'
         # httplib explodes if url is unicode
-        self.server = xmlrpclib.ServerProxy(str(url + '/xmlrpc.cgi'))
+        self.server = xmlrpclib.ServerProxy(str(url))
 
     def rpc(self, *args, **kwargs):
         """Do an RPC on the Bugzilla server.
