@@ -17,7 +17,9 @@
 from __future__ import unicode_literals
 
 import argparse
+import datetime
 import itertools
+import re
 import textwrap
 
 from . import bug
@@ -25,12 +27,24 @@ from . import bugzilla
 from . import config
 from . import editor
 
+conf = config.Config.get_config('~/.bugzillarc')
+
 
 class _ReadFileAction(argparse.Action):
     def __call__(self, parser, namespace, values, option_string):
         setattr(namespace, self.dest, values.read())
 
 
+def date(s):
+    match = re.match(r'(\d{4})-(\d\d)-(\d\d)$', s)
+    if not match:
+        raise argparse.ArgumentTypeError('Date must be in format: YYYY-MM-DD')
+    try:
+        return datetime.date(*map(int, match.group(1, 2, 3)))
+    except ValueError as e:
+        raise argparse.ArgumentTypeError(e.message)
+
+
 def with_add_remove(src, dst, metavar=None, type=None):
     def decorator(cls):
         cls.args = cls.args + [
@@ -64,6 +78,22 @@ def with_bugs(cls):
     return cls
 
 
+def with_time(cls):
+    cls.args = cls.args + [
+        lambda x: x.add_argument('--estimated-time', type=float,
+            help='Estimate of the total time required to fix the bug, '
+                 'in hours.'),
+        lambda x: x.add_argument('--work-time', type=float,
+            help='Additional hours worked.'),
+        lambda x: x.add_argument('--remaining-time', type=float,
+            help='Estimated time remaining.  If not supplied, any hours '
+                 'worked will be deducted from the current remaining time.'),
+        lambda x: x.add_argument('--deadline', type=date,
+            help='Date when the bug must be fixed, in format YYYY-MM-DD'),
+    ]
+    return cls
+
+
 def with_optional_message(cls):
     def msgargs(parser):
         group = parser.add_mutually_exclusive_group()
@@ -87,6 +117,21 @@ def with_limit(things='items', default=None):
     return decorator
 
 
+def with_server(cls):
+    def add_server_args(parser):
+        default = None
+        if conf.has_option('core', 'server'):
+            default = conf.get('core', 'server')
+        group = parser.add_argument_group('server arguments')
+        group.add_argument('--server', default=default,
+            help='Handle of Bugzilla instance to use')
+        group.add_argument('--url', help='Base URL of Bugzilla instance')
+        group.add_argument('--user', help='Bugzilla username')
+        group.add_argument('--password', help='Bugzilla password')
+    cls.args = cls.args + [add_server_args]
+    return cls
+
+
 class Command(object):
     """A command object.
 
@@ -121,9 +166,58 @@ class Command(object):
         self._ui = ui
 
 
+class Config(Command):
+    """Show or update configuration."""
+    args = Command.args + [
+        lambda x: x.add_argument('--list', '-l', action='store_true',
+            help='list all configuration options'),
+        lambda x: x.add_argument('name', nargs='?',
+            help='name of option to show, set or remove'),
+        lambda x: x.add_argument('--remove', action='store_true',
+            help='remove the specified option'),
+        lambda x: x.add_argument('value', nargs='?',
+            help='set value of given option'),
+    ]
+
+    def __call__(self):
+        args = self._args
+        if args.list:
+            for section in conf.sections():
+                for option, value in conf.items(section):
+                    print '{}={}'.format('.'.join((section, option)), value)
+        elif not args.name:
+            raise UserWarning('No configuration option given.')
+        else:
+            try:
+                section, option = args.name.rsplit('.', 1)
+            except ValueError:
+                raise UserWarning('Invalid configuration option.')
+            if not section or not option:
+                raise UserWarning('Invalid configuration option.')
+
+            if args.remove:
+                # remove the option
+                conf.remove_option(section, option)
+                if not conf.items(section):
+                    conf.remove_section(section)
+                conf.write()
+            elif args.value:
+                # set new value
+                if not conf.has_section(section):
+                    conf.add_section(section)
+                oldvalue = conf.get(section, option) \
+                    if conf.has_option(section, option) else None
+                conf.set(section, option, args.value)
+                conf.write()
+                print '{}: {} => {}'.format(args.name, oldvalue, args.value)
+            else:
+                curvalue = conf.get(section, option)
+                print '{}: {}'.format(args.name, curvalue)
+
+
 class Help(Command):
     """Show help."""
-    args = [
+    args = Command.args + [
         lambda x: x.add_argument('subcommand', metavar='SUBCOMMAND', nargs='?',
             help='show help for subcommand')
     ]
@@ -143,35 +237,34 @@ class Help(Command):
                 self._parser.parse_args([self._args.subcommand, '--help'])
 
 
+@with_server
 class BugzillaCommand(Command):
     def __init__(self, *args, **kwargs):
         super(BugzillaCommand, self).__init__(*args, **kwargs)
         # construct the Bugzilla object
         args = self._args
-        url, user, password = args.url, args.user, args.password
-        if not (url and user and password):
-            if not args.server:
-                raise UserWarning("No server specified.")
+        server = {}
+        if args.server:
             try:
-                server = config.get('servers').get(args.server)
-            except AttributeError:
-                raise UserWarning("No servers defined.")
-            if not server:
+                server = dict(conf.items('server.' + args.server))
+            except config.NoSectionError:
                 raise UserWarning(
                     "No configuration for server '{}'.".format(args.server)
                 )
-            url = url or server[0]
-            user = user or server[1]
-            password = password or server[2]
-        self.bz = bugzilla.Bugzilla(url, user, password)
+        server.update({
+            k: getattr(args, k)
+            for k in ('url', 'user', 'password')
+            if getattr(args, k)
+        })
+        self.bz = bugzilla.Bugzilla(**server)
 
 
 @with_bugs
 @with_optional_message
 class Assign(BugzillaCommand):
     """Assign bugs to the given user."""
-    args = [
-        lambda x: x.add_argument('--to', metavar='ASSIGNEE',
+    args = BugzillaCommand.args + [
+        lambda x: x.add_argument('--to', metavar='ASSIGNEE', required=True,
             help='New assignee'),
     ]
 
@@ -210,7 +303,6 @@ class Block(BugzillaCommand):
         else:
             # show blocked bugs
             for bug in bugs:
-                bug.read()
                 print 'Bug {}:'.format(bug.bugno)
                 if bug.data['blocks']:
                     print '  Blocked bugs: {}'.format(
@@ -249,7 +341,6 @@ class CC(BugzillaCommand):
         else:
             # show CC List
             for bug in bugs:
-                bug.read()
                 print 'Bug {}:'.format(bug.bugno)
                 if bug.data['cc']:
                     print '  CC List: {}'.format(
@@ -263,13 +354,19 @@ class CC(BugzillaCommand):
 @with_limit(things='comments')
 class Comment(BugzillaCommand):
     """List comments or file a comment on the given bugs."""
-    args = [
+    args = BugzillaCommand.args + [
         lambda x: x.add_argument('--reverse', action='store_true',
             default=True,
-            help='Show from newest to oldest.'),
+            help='Show from newest to oldest (the default).'),
         lambda x: x.add_argument('--forward', action='store_false',
             dest='reverse',
             help='Show from oldest to newest.'),
+        lambda x: x.add_argument('--omit-empty', action='store_true',
+            default=True,
+            help='Omit empty comments (the default).'),
+        lambda x: x.add_argument('--include-empty', action='store_false',
+            dest='omit_empty',
+            help='Include empty comments.'),
     ]
 
     formatstring = '{}\nauthor: {creator}\ntime: {time}\n\n{text}\n\n'
@@ -283,7 +380,7 @@ class Comment(BugzillaCommand):
         else:
             def cmtfmt(bug):
                 comments = sorted(
-                    enumerate(self.bz.bug(bug).get_comments()),
+                    enumerate(self.bz.bug(bug).comments),
                     key=lambda x: int(x[1]['id']),
                     reverse=True  # initially reverse to apply limit
                 )
@@ -297,13 +394,13 @@ class Comment(BugzillaCommand):
 
                 return '=====\nBUG {}\n\n-----\n{}'.format(
                     bug,
-                    '-----\n'.join(map(
-                        lambda (n, x): self.formatstring.format(
+                    '-----\n'.join(
+                        self.formatstring.format(
                             'comment: {}'.format(n) if n else 'description',
-                            **x
-                        ),
-                        comments
-                    ))
+                            **comment)
+                        for n, comment in comments
+                        if not args.omit_empty or comment['text']
+                    )
                 )
             print '\n'.join(map(cmtfmt, args.bugs))
 
@@ -333,7 +430,6 @@ class Depend(BugzillaCommand):
         else:
             # show dependencies
             for bug in bugs:
-                bug.read()
                 print 'Bug {}:'.format(bug.bugno)
                 if bug.data['depends_on']:
                     print '  Dependencies: {}'.format(
@@ -342,6 +438,29 @@ class Depend(BugzillaCommand):
                     print '  No dependencies'
 
 
+@with_bugs
+class Desc(BugzillaCommand):
+    """Show the description of the given bug(s)."""
+    formatstring = 'author: {creator}\ntime: {time}\n\n{text}\n'
+
+    def __call__(self):
+        def _descfmt(bug):
+            desc = self.bz.bug(bug).comments[0]
+            return '=====\nBUG {}\n{}'.format(
+                bug,
+                self.formatstring.format(**desc)
+            )
+        print '\n'.join(_descfmt(bug) for bug in self._args.bugs)
+
+
+@with_bugs
+class Dump(BugzillaCommand):
+    """Print internal representation of bug data."""
+    def __call__(self):
+        bugs = (self.bz.bug(x) for x in self._args.bugs)
+        print '\n'.join(str((x.data, x.comments)) for x in bugs)
+
+
 class Fields(BugzillaCommand):
     """List valid values for bug fields."""
     def __call__(self):
@@ -370,11 +489,10 @@ class Info(BugzillaCommand):
     """Show detailed information about the given bugs."""
     def __call__(self):
         args = self._args
-        fields = config.get_show_fields()
+        fields = config.show_fields
         for bug in map(self.bz.bug, args.bugs):
-            bug.read()
             print 'Bug {}:'.format(bug.bugno)
-            fields = config.get_show_fields() & bug.data.viewkeys()
+            fields = config.show_fields & bug.data.viewkeys()
             width = max(map(len, fields)) - min(map(len, fields)) + 2
             for field in fields:
                 print '  {:{}} {}'.format(field + ':', width, bug.data[field])
@@ -386,11 +504,9 @@ class List(BugzillaCommand):
     """Show a one-line summary of the given bugs."""
     def __call__(self):
         args = self._args
-        fields = config.get_show_fields()
         lens = map(lambda x: len(str(x)), args.bugs)
         width = max(lens) - min(lens) + 2
         for bug in map(self.bz.bug, args.bugs):
-            bug.read()
             print 'Bug {:{}} {}'.format(
                 str(bug.bugno) + ':', width, bug.data['summary']
             )
@@ -523,7 +639,7 @@ class Status(BugzillaCommand):
     status and resolution fields to appropriate values for duplicate bugs.
     """
 
-    args = [
+    args = BugzillaCommand.args + [
         lambda x: x.add_argument('--status',
             help='Specify a resolution (case-insensitive).'),
         lambda x: x.add_argument('--resolution',
@@ -597,6 +713,52 @@ class Status(BugzillaCommand):
 #    pass
 
 
+@with_bugs
+@with_optional_message
+@with_time
+class Time(BugzillaCommand):
+    """Show or adjust times and estimates for the given bugs."""
+    def __call__(self):
+        args = self._args
+
+        message = editor.input('Enter your comment.') if args.message is True \
+            else args.message
+
+        time_args = \
+            ['estimated_time', 'remaining_time', 'work_time', 'deadline']
+        if any(getattr(args, arg) is not None for arg in time_args):
+            # adjust
+            if len(args.bugs) != 1:
+                # makes no sense to adjust times on several bugs at once
+                raise UserWarning('Cannot adjust times on multiple bugs.')
+            self.bz.bug(args.bugs[0]).update(
+                estimated_time=args.estimated_time,
+                remaining_time=args.remaining_time,
+                work_time=args.work_time,
+                deadline=args.deadline,
+                comment=message
+            )
+        else:
+            # display
+            #
+            # as of Bugzilla 4.0.1, "total hours worked" cannot be known.
+            # Hours are accumulated in the 'work_time' attribute of comments,
+            # which is not present in bug.comments RPC output.
+            bugs = (self.bz.bug(bug) for bug in args.bugs)
+            for bug in bugs:
+                # if user is not in the "time-tracking" group, the fields will
+                # be absent from bug data.  first check that they're there.
+                time_fields = ('deadline', 'estimated_time', 'remaining_time')
+                if not all(x in bug.data for x in time_fields):
+                    print 'User is not in the time-tracking group.'
+                    return
+                print 'Bug {}:'.format(bug.bugno)
+                print '  Estimated time: {}'.format(bug.data['estimated_time'])
+                print '  Remaining time: {}'.format(bug.data['remaining_time'])
+                print '  Deadline:       {}'.format(bug.data['deadline'])
+                print '  Time worked:    {}'.format('<not available>')
+
+
 # the list got too long; metaprogram it ^_^
 commands = filter(
     lambda x: type(x) == type                     # is a class \
